package de.fitko.fitconnect.api.services.submission;

import de.fitko.fitconnect.api.domain.model.destination.Destination;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.*;

import java.util.UUID;

/**
 *  A service that provides access to the FIT-Connect REST-API and allows creating and loading {@link Submission}s
 *  and {@link Attachment}s.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/apis/submission-api">FIT-Connect Submission API</a>
 */
public interface SubmissionService {

    /**
     * Announce a new submission. This step is necessary before actually {@link #sendSubmission(SubmitSubmission) sending} the submission.
     *
     * @param submission submission that contains all {@link Attachment}s that should be announced, as well as a specified {@link ServiceType}
     * @return announced submission with submissionId, caseId and destinationID
     *
     * @see <a href="https://docs.fitko.de/fit-connect/docs/sending/start-submission">Announcing a submission</a>
     */
    SubmissionForPickup announceSubmission(CreateSubmission submission);

    /**
     * Send a submission that was already {@link #announceSubmission(CreateSubmission) announced} before.
     *
     * @param submission submission including the encrypted {@link Data} and {@link Metadata}
     * @return the submission that was sent
     */
    Submission sendSubmission(SubmitSubmission submission);

    /**
     * Get a {@link Submission} by id.
     *
     * @param submissionId unique submission identifier
     * @return submission matching the given id
     */
    Submission getSubmission(UUID submissionId);

    /**
     * Get all available submissions for a given {@link Destination}.
     *
     * @param destinationId unique destination identifier
     * @param offset position in the dataset
     * @param limit number of submissions in result (max. is 500)
     * @return SubmissionsForPickup containing a list of submissions
     */
    SubmissionsForPickup pollAvailableSubmissions(UUID destinationId, int offset, int limit);

    /**
     * Upload an encrypted {@link Attachment}.
     *
     * @param submissionId unique identifier of an announced submission
     * @param attachmentId unique destination identifier
     * @param encryptedAttachment JWE encrypted attachment payload
     */
    void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment);

    /**
     * Get an {@link Attachment} by id for a given {@link Submission}.
     *
     * @param submissionId unique submission identifier
     * @param attachmentId unique attachment identifier
     * @return encrypted string of the attachment data
     */
    String getAttachment(UUID submissionId, UUID attachmentId);
}
