package de.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.destination.Destination;
import de.fitko.fitconnect.api.domain.model.destination.DestinationService;
import de.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.signature.AttachmentSignatureType;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.signature.Hash__1;
import de.fitko.fitconnect.api.domain.model.metadata.data.*;
import de.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import de.fitko.fitconnect.api.domain.model.submission.ServiceType;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.*;
import de.fitko.fitconnect.api.services.Sender;
import lombok.Builder;
import lombok.Getter;
import lombok.With;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.*;
import java.util.stream.Collectors;

/**
 * A fluent client for announcing and handing in a {@link SubmitSubmission}
 */
public class SenderClient {

    private static final Logger logger = LoggerFactory.getLogger(SenderClient.class);

    private SenderClient() {
    }

    public static WithDestination build(final Sender sender) {
        return new ClientBuilder(sender);
    }

    public interface WithDestination {

        /**
         * Configures the client for the given destination and loads the public key
         *
         * @param destinationId unique identifier of the clients destination
         * @return the upload step for attachments
         */
        WithServiceType withDestination(UUID destinationId);
    }

    public interface WithAttachments {

        /**
         * Sends the submission with a list of attachments
         *
         * @param attachments that are sent with the submission
         * @return the next step where additional data can be added to the submission
         */
        WithData withAttachments(List<File> attachments);

        /**
         * Sends the submission with an attachments
         *
         * @param attachment that is sent with the submission
         * @return the next step where additional data can be added to the submission
         */
        WithData withAttachment(File attachment);

        /**
         * XML data as string.
         *
         * @param data json string
         * @return next step to submit the data
         */
        Submit withJsonData(String data);

        /**
         * JSON data as string.
         *
         * @param data xml string
         * @return next step to submit the data
         */
        Submit withXmlData(String data);
    }

    public interface WithServiceType {

        /**
         * Sets a {@link ServiceType} for a submission
         *
         * @param name     a name of the service
         * @param leikaKey unique identifier of the service in the form <b>urn:de:fim:leika:leistung:</b>
         * @return next step to add attachments
         * @see <a href="https://fimportal.de/">Search For Service Types</a>
         */
        WithAttachments withServiceType(String name, String leikaKey);

        /**
         * Sets a {@link ServiceType} for a submission
         *
         * @param name        a name of the service
         * @param description description of the service
         * @param leikaKey    unique identifier of the service in the form <b>urn:de:fim:leika:leistung:</b>
         * @return next step to add attachments
         * @see <a href="https://fimportal.de/">Search For Service Types</a>
         */
        WithAttachments withServiceType(String name, String description, String leikaKey);
    }

    public interface WithData {

        /**
         * XML data as string.
         *
         * @param data json string
         * @return next step to submit the data
         */
        Submit withJsonData(String data);

        /**
         * JSON data as string.
         *
         * @param data xml string
         * @return next step to submit the data
         */
        Submit withXmlData(String data);

    }

    public interface Submit {

        /**
         * Send submission with attachments and data to FIT-Connect API.
         *
         * @return {@link SubmitSubmission} object of the handed in submission
         */
        Optional<SubmitSubmission> submit();
    }

    /**
     * ClientBuilder that implements all steps to guide through the API calls
     */
    public static class ClientBuilder implements WithDestination, WithServiceType, WithAttachments, WithData, Submit {

        private final Sender sender;

        private DataPayload dataPayload;
        private UUID destinationId;
        private ServiceType serviceType;
        private List<File> attachments = new ArrayList<>();

        public ClientBuilder(final Sender sender) {
            this.sender = sender;
        }

        @Override
        public WithServiceType withDestination(final UUID destinationId) {
            this.destinationId = destinationId;
            return this;
        }

        @Override
        public WithAttachments withServiceType(final String name, final String leikaKey) {
            this.serviceType = createServiceTypeWithoutDescription(name, leikaKey);
            return this;
        }

        @Override
        public WithAttachments withServiceType(final String name, final String description, final String leikaKey) {
            this.serviceType = createServiceTypeWithDescription(name, description, leikaKey);
            return this;
        }

        @Override
        public Submit withXmlData(final String data) {
            this.dataPayload = getDataPayload(data, MimeType.APPLICATION_XML);
            return this;
        }

        @Override
        public Submit withJsonData(final String data) {
            this.dataPayload = getDataPayload(data, MimeType.APPLICATION_JSON);
            return this;
        }

        @Override
        public WithData withAttachment(final File attachmentFile) {
            this.attachments = attachmentFile != null ? List.of(attachmentFile) : List.of();
            return this;
        }

        @Override
        public WithData withAttachments(final List<File> attachmentFiles) {
            this.attachments = attachmentFiles;
            return this;
        }

        @Override
        public Optional<SubmitSubmission> submit() {
            try {

                if (!isAllNecessaryDataSet()) return Optional.empty();

                /** Get encryption key for destination **/
                final Destination destination = sender.getDestination(destinationId);
                final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId, destination.getEncryptionKid());
                final ValidationResult validationResult = sender.validatePublicKey(encryptionKey);
                if (validationResult.hasError()) {
                    logger.info("The public key is not valid", validationResult.getError());
                    return Optional.empty();
                }

                /** Create new submission and announce attachments **/
                final List<AttachmentPayload> attachmentPayloads = encryptAndHashFiles(encryptionKey, attachments);
                final List<UUID> attachmentIdsToAnnounce = toAttachmentIds(attachmentPayloads);
                final CreateSubmission newSubmission = createSubmission(attachmentIdsToAnnounce);
                final SubmissionForPickup announcedSubmission = sender.createSubmission(newSubmission);
                final UUID submissionId = announcedSubmission.getSubmissionId();
                uploadAttachments(attachmentPayloads, submissionId);

                /** Prepare submit submission with announced submission id **/
                final SubmitSubmission submission = new SubmitSubmission();
                submission.setSubmissionId(submissionId);

                /** Create attachment hashes **/
                final List<Attachment> hashedAttachments = toHashedAttachments(attachmentPayloads);

                /** Set encrypted metadata with data payload and attachments  **/
                logger.info("Adding data payload with mime-type {} to submission", this.dataPayload.mimeType);
                final DataPayload data = getEncryptedData(this.dataPayload, destination, encryptionKey);
                submission.setEncryptedData(data.getEncryptedData());
                submission.setEncryptedMetadata(getEncryptedMetadata(encryptionKey, data, hashedAttachments));

                /** submit submission **/
                sender.sendSubmission(submission);
                logger.info("SUCCESSFULLY HANDED IN SUBMISSION !");
                return Optional.of(submission);

            } catch (final EncryptionException e) {
                logger.error("Encrypting submission failed", e);
            } catch (final RestApiException e) {
                logger.error("Sending submission failed", e);
            } catch (final SchemaNotFoundException e) {
                logger.error("Required schema to send valid submission not found", e);
            } catch (final SubmissionNotCreatedException e) {
                logger.error("Failed to announce new submission", e);
            } catch (final KeyNotRetrievedException e) {
                logger.error("Getting encryption key for destination {} failed", destinationId, e);
            } catch (final AttachmentCreationException e) {
                logger.error("Reading file failed. Attachment will not be created.", e);
            }
            return Optional.empty();
        }

        private DataPayload getEncryptedData(final DataPayload dataPayload, final Destination destination, final RSAKey encryptionKey) {
            final List<SubmissionSchema> submissionSchemas = getSubmissionSchemasFromDestination(destination);
            final Optional<URI> schemaUriForMimeType = getSchemaUriForMimeType(submissionSchemas, dataPayload.mimeType);
            if (schemaUriForMimeType.isEmpty()) {
                throw new SchemaNotFoundException("No data schema for mime-type " + dataPayload.mimeType + " found, please check the allowed type for destination");
            }
            final String hashedData = sender.createHash(dataPayload.rawData);
            final String encryptedData = sender.encryptBytes(encryptionKey, dataPayload.rawData);
            return dataPayload.withSchemaUri(schemaUriForMimeType.get())
                    .withEncryptedData(encryptedData)
                    .withHashedData(hashedData);
        }

        private String getEncryptedMetadata(final RSAKey encryptionKey, final DataPayload dataPayload, final List<Attachment> hashedAttachments) {
            final Metadata metadata = createMetadata(hashedAttachments, dataPayload);
            return sender.encryptObject(encryptionKey, metadata);
        }

        private List<Attachment> toHashedAttachments(final List<AttachmentPayload> attachmentPayloads) {
            return attachmentPayloads.stream()
                    .map(this::toHashedAttachment)
                    .collect(Collectors.toList());
        }

        private CreateSubmission createSubmission(final List<UUID> attachmentIdsToAnnounce) {
            return CreateSubmission.builder()
                    .destinationId(destinationId)
                    .announcedAttachments(attachmentIdsToAnnounce)
                    .serviceType(serviceType)
                    .build();
        }

        private List<UUID> toAttachmentIds(final List<AttachmentPayload> attachmentPayloads) {
            return attachmentPayloads.stream()
                    .map(AttachmentPayload::getAttachmentId)
                    .collect(Collectors.toList());
        }

        private DataPayload getDataPayload(final String data, final MimeType mimeType) {
            if (data == null) {
                return null;
            }
            return DataPayload.builder()
                    .rawData(data.getBytes(StandardCharsets.UTF_8))
                    .mimeType(mimeType)
                    .build();
        }

        private List<SubmissionSchema> getSubmissionSchemasFromDestination(final Destination destination) {
            return destination.getServices().stream()
                    .map(DestinationService::getSubmissionSchemas)
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
        }

        private Optional<URI> getSchemaUriForMimeType(final List<SubmissionSchema> submissionSchemas, final MimeType mimeType) {
            return submissionSchemas.stream()
                    .filter(schema -> schema.getMimeType().equals(mimeType))
                    .map(SubmissionSchema::getSchemaUri)
                    .findFirst();
        }

        private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
            if (attachmentPayloads.isEmpty()) {
                logger.info("No attachments to upload");
            } else {
                logger.info("Uploading {} attachment(s)", attachmentPayloads.size());
            }
            for (final AttachmentPayload payload : attachmentPayloads) {
                sender.uploadAttachment(submissionId, payload.attachmentId, payload.encryptedData);
            }
        }

        private ServiceType createServiceTypeWithDescription(final String name, final String description, final String leikaKey) {
            final var serviceType = new ServiceType();
            serviceType.setName(name);
            serviceType.setDescription(description);
            serviceType.setIdentifier(leikaKey);
            return serviceType;
        }

        private ServiceType createServiceTypeWithoutDescription(final String name, final String leikaKey) {
            final var serviceType = new ServiceType();
            serviceType.setName(name);
            serviceType.setIdentifier(leikaKey);
            return serviceType;
        }

        private Data createData(final DataPayload dataPayload) {
            final var hash = new Hash();
            hash.setContent(dataPayload.hashedData);
            hash.setDataSignatureType(DataSignatureType.SHA_512);

            final var submissionSchema = new SubmissionSchema();
            submissionSchema.setMimeType(dataPayload.mimeType);
            submissionSchema.setSchemaUri(dataPayload.schemaUri);

            final var data = new Data();
            data.setSubmissionSchema(submissionSchema);
            data.setHash(hash);
            return data;
        }

        private Metadata createMetadata(final List<Attachment> attachments, final DataPayload dataPayload) {
            final var contentStructure = new ContentStructure();
            contentStructure.setAttachments(attachments);
            contentStructure.setData(createData(dataPayload));

            final var metadata = new Metadata();
            metadata.setContentStructure(contentStructure);
            metadata.setPublicServiceType(getPublicServiceType());
            return metadata;
        }

        private PublicServiceType getPublicServiceType() {
            final var publicServiceType = new PublicServiceType();
            publicServiceType.setIdentifier(serviceType.getIdentifier());
            publicServiceType.setName(serviceType.getName());
            if (serviceType.getDescription() != null) {
                publicServiceType.setDescription(serviceType.getDescription());
            }
            return publicServiceType;
        }

        private List<AttachmentPayload> encryptAndHashFiles(final RSAKey encryptionKey, final List<File> attachmentFiles) {
            return attachmentFiles.stream()
                    .map(this::readRawData)
                    .map(this::hashBytes)
                    .map(payload -> encryptBytes(encryptionKey, payload))
                    .map(this::setMimeType)
                    .collect(Collectors.toList());
        }

        private AttachmentPayload readRawData(final File file) {
            try {
                final byte[] rawData = Files.readAllBytes(Paths.get(file.getPath()));
                return AttachmentPayload.builder().file(file).rawData(rawData).build();
            } catch (final IOException e) {
                throw new AttachmentCreationException("Attachment '" + file.getAbsolutePath() + "' could not be read ", e);
            }
        }

        private Attachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
            final var attachment = new Attachment();
            attachment.setAttachmentId(attachmentPayload.attachmentId);
            attachment.setPurpose(Purpose.ATTACHMENT);
            attachment.setFilename(attachmentPayload.file.getName());
            attachment.setMimeType(attachmentPayload.mimeType);

            final var hash = new Hash__1();
            hash.setContent(attachmentPayload.hashedData);
            hash.setAttachmentSignatureType(AttachmentSignatureType.SHA_512);
            attachment.setHash(hash);
            return attachment;
        }

        private AttachmentPayload encryptBytes(final RSAKey encryptionKey, final AttachmentPayload attachmentPayload) {
            final String encryptedAttachment = sender.encryptBytes(encryptionKey, attachmentPayload.rawData);
            return attachmentPayload.withEncryptedData(encryptedAttachment);
        }

        private AttachmentPayload hashBytes(final AttachmentPayload attachmentPayload) {
            final String hashedBytes = sender.createHash(attachmentPayload.rawData);
            return attachmentPayload.withHashedData(hashedBytes);
        }

        private AttachmentPayload setMimeType(final AttachmentPayload attachmentPayload) {
            final File file = attachmentPayload.file;
            String mimeType;
            try {
                final Tika mimeTypeGuesser = new Tika();
                mimeType = mimeTypeGuesser.detect(file);
            } catch (final IOException e) {
                mimeType = URLConnection.guessContentTypeFromName(file.getName());
            }
            logger.info("Detected attachment mime-type {}", mimeType);
            return attachmentPayload.withMimeType(mimeType);
        }

        private boolean isAllNecessaryDataSet() {
            if (this.dataPayload == null) {
                logger.error("Data is mandatory, but was null.");
                return false;
            } else if (this.serviceType.getName() == null) {
                logger.error("Service type name is mandatory, but was null.");
                return false;
            } else if (this.serviceType.getIdentifier() == null) {
                logger.error("Service type identifier is mandatory, but was null.");
                return false;
            } else if (this.destinationId == null) {
                logger.error("DestinationId is mandatory, but was null.");
                return false;
            } else {
                return true;
            }
        }

        @With
        @Getter
        @Builder
        private static class AttachmentPayload {
            private File file;
            private byte[] rawData;
            private String hashedData;
            private String encryptedData;
            private String mimeType;
            private final UUID attachmentId = UUID.randomUUID();
        }

        @With
        @Getter
        @Builder
        private static class DataPayload {
            private byte[] rawData;
            private String hashedData;
            private String encryptedData;
            private MimeType mimeType;
            private URI schemaUri;
        }
    }
}