package de.fitko.fitconnect.client;

import de.fitko.fitconnect.api.config.*;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.client.factory.ClientFactory;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;

import java.io.File;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import static de.fitko.fitconnect.api.config.UsedEnvironment.TEST;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * This test uses real credentials and endpoints stored in gitlab ci variables to test the sdk against a real system
 */
public class ClientIntegrationTest {

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    public void testSendAndConfirmCycle() {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab();

        final var sentSubmission = ClientFactory.senderClient(config)
                .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{ data: 'Beispiel Fachdaten' }")
                .submit();

        assertTrue(sentSubmission.isPresent());

        // When
        final Optional<SubscriberClient.ReceivedSubmission> receivedSubmission =
                ClientFactory.subscriberClient(config)
                        .requestSubmission(sentSubmission.get().getSubmissionId());

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is("{ data: 'Beispiel Fachdaten' }"));
        assertThat(receivedSubmission.get().getMimeType().toString(), is("application/json"));
        assertThat(new String(receivedSubmission.get().getAttachments().get(0).getDecryptedData()), is("Test attachment"));
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    public void testListSubmissions() {

        // Given
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab();
        final var senderClient = ClientFactory.senderClient(config);
        final var subscriberClient = ClientFactory.subscriberClient(config);

        final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
        final String leikaKey = "urn:de:fim:leika:leistung:99400048079000";
        final String serviceName = "Test Service";

        final var submissionOne = senderClient
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .withJsonData("{ data: 'Beispiel Fachdaten 1' }")
                .submit();

        final var submissionTwo = senderClient
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .withJsonData("{ data: 'Beispiel Fachdaten 2' }")
                .submit();

        assertTrue(submissionOne.isPresent());
        assertTrue(submissionTwo.isPresent());

        // When
        final Set<SubmissionForPickup> submissions = subscriberClient.getAvailableSubmissions(destinationId);

        // Then
        assertFalse(submissions.isEmpty());

        final List<UUID> submissionIds = submissions.stream().map(SubmissionForPickup::getSubmissionId).collect(Collectors.toList());

        assertThat(submissionIds, hasItems(submissionOne.get().getSubmissionId(), submissionTwo.get().getSubmissionId()));

        // remove by confirming
        subscriberClient.requestSubmission(submissionOne.get().getSubmissionId());
        subscriberClient.requestSubmission(submissionTwo.get().getSubmissionId());
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab() {

        final var testEnvironment = new Environment();
        testEnvironment.setAuthBaseUrl("https://auth-testing.fit-connect.fitko.dev");
        testEnvironment.setSubmissionBaseUrl("https://submission-api-testing.fit-connect.fitko.dev");

        final var environments = new Environments();
        environments.setTest(testEnvironment);

        final var sender = new Sender();
        sender.setClientId(System.getenv("SENDER_CLIENT_ID"));
        sender.setClientSecret(System.getenv("SENDER_CLIENT_SECRET"));

        final var subscriber = new Subscriber();
        subscriber.setClientId(System.getenv("SUBSCRIBER_CLIENT_ID"));
        subscriber.setClientSecret(System.getenv("SUBSCRIBER_CLIENT_SECRET"));
        subscriber.setPrivateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json");
        subscriber.setPrivateSigningKeyPath("src/test/resources/private_test_signing_key.json");
        subscriber.setMetadataSchemaPath("src/test/resources/metadata_schema.json");

        final var resourcePaths = new ResourcePaths();
        resourcePaths.setAuthTokenPath("/token");
        resourcePaths.setDestinationPath("/v1/destinations/{destinationId}");
        resourcePaths.setDestinationKeyPath("/v1/destinations/{destinationId}/keys/{kid}");
        resourcePaths.setEventLogPath("/v1/cases/{caseId}/events");
        resourcePaths.setSubmissionsPath("/v1/submissions");
        resourcePaths.setSubmissionPath("/v1/submissions/{submissionId}");
        resourcePaths.setSubmissionAttachmentPath("/v1/submissions/{submissionId}/attachments/{attachmentId}");

        final var config = new ApplicationConfig();
        config.setSender(sender);
        config.setSubscriber(subscriber);
        config.setEnvironments(environments);
        config.setUsedEnvironment(TEST);
        config.setResourcePaths(resourcePaths);

        return config;
    }
}
