package de.fitko.fitconnect.client;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import de.fitko.fitconnect.api.domain.model.destination.Destination;
import de.fitko.fitconnect.api.domain.model.destination.DestinationService;
import de.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import de.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.EncryptionException;
import de.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import de.fitko.fitconnect.api.exceptions.RestApiException;
import de.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import de.fitko.fitconnect.api.services.Sender;
import io.github.netmikey.logunit.api.LogCapturer;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.RegisterExtension;
import org.mockito.Mockito;

import java.io.File;
import java.net.URI;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

public class SenderClientTest {

    @RegisterExtension
    LogCapturer logs = LogCapturer.create().captureForType(SenderClient.class);

    private Sender senderMock;
    private SenderClient.WithDestination underTest;

    @BeforeEach
    public void setup() {
        senderMock = Mockito.mock(Sender.class);
        underTest = SenderClient.build(senderMock);
    }

    @Test
    public void testSendSubmissionWithJsonData() throws Exception {

        // Given

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);
        final SubmissionSchema jsonSchema = new SubmissionSchema();
        jsonSchema.setSchemaUri(URI.create("https://json.mimetype"));
        jsonSchema.setMimeType(MimeType.APPLICATION_JSON);
        final Destination destination = getDestination(destinationId, jsonSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(announcedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    public void testSendSubmissionWithXmlData() throws Exception {

        // Given

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);
        final SubmissionSchema xmlSchema = new SubmissionSchema();
        xmlSchema.setSchemaUri(URI.create("https://xml.mimetype"));
        xmlSchema.setMimeType(MimeType.APPLICATION_XML);
        final Destination destination = getDestination(destinationId, xmlSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withXmlData("<xml><test>data</test></xml>")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        assertThat(submission.get().getSubmissionId(), equalTo(announcedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    public void testMimeTypeDetection() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();
        final File file1 = new File("src/test/resources/text_attachment_with_no_filetype");
        final File file2 = new File("src/test/resources/pdf_attachment_with_no_extension");

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withAttachments(List.of(file1, file2))
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("Detected attachment mime-type text/plain");
        logs.assertContains("Detected attachment mime-type application/pdf");
    }


    @Test
    public void testWithMissingData() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData(null)
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Data is mandatory, but was null");
    }

    @Test
    public void testWithMissingDestinationId() throws Exception {

        // Given
        setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(null)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("DestinationId is mandatory, but was null");
    }

    @Test
    public void testMissingDataMimeTypeSchema() throws Exception {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("Service", "de:fitco:test:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Required schema to send valid submission not found");
    }

    @Test
    public void testWithMissingServiceTypeName() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType(null, "leika:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Service type name is mandatory, but was null");
    }

    @Test
    public void testWithOptionalServiceTypeIdentifier() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("Name", null)
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Service type identifier is mandatory, but was null");
    }

    @Test
    public void testWithMissingServiceTypeDescription() throws Exception {

        // Given
        final var destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("Name", null, "leika:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isPresent());
    }

    @Test
    public void testWithEmptyAttachments() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withAttachments(List.of())
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    public void testWithAttachmentFileNull() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withAttachment(null)
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isPresent());
        logs.assertContains("No attachments to upload");
    }

    @Test
    public void testWithAttachmentPathNotPresent() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withAttachment(new File("/non/existing/path"))
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Reading file failed. Attachment will not be created.");
    }

    @Test
    public void testInvalidPublicKey() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.error(new Exception()));

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("The public key is not valid");
    }

    @Test
    public void testEncryptionError() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.encryptBytes(any(),any())).thenThrow(new EncryptionException("Encryption failed"));

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Encrypting submission failed");
    }

    @Test
    public void testAnnouncingSubmissionFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.createSubmission(any())).thenThrow(new SubmissionNotCreatedException("Announcing submission failed"));

        // When
        final var submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        // Then
        assertTrue(submission.isEmpty());
        logs.assertContains("Failed to announce new submission");

    }

    @Test
    public void testRestApiCallFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.getDestination(any())).thenThrow(new RestApiException("Loading destination failed"));

        // When
        final Optional<SubmitSubmission> submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Sending submission failed");

    }

    @Test
    public void testGettingEncryptionKeyFailed() throws Exception {

        // Given
        final var destinationId = setupTestMocks();
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenThrow(new KeyNotRetrievedException("Getting encryption key failed"));

        // When
        final Optional<SubmitSubmission> submission = underTest
                .withDestination(destinationId)
                .withServiceType("name", "test:key")
                .withJsonData("{}")
                .submit();

        assertTrue(submission.isEmpty());
        logs.assertContains("Getting encryption key for destination "+ destinationId+" failed");
    }

    private UUID setupTestMocks() throws JOSEException {
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = generateRsaKey(4096).toPublicJWK();

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any(), any())).thenReturn(publicKey);
        return destinationId;
    }

    private SubmissionForPickup getAnnouncedSubmission(final UUID destinationId) {
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://json-schema.fitco.de"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("1");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));

        final var announcedSubmission = new SubmissionForPickup();
        announcedSubmission.setSubmissionId(UUID.randomUUID());
        announcedSubmission.setDestinationId(destinationId);
        return announcedSubmission;
    }

    private Destination getDestination(final UUID destinationId) {
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://mimetype.test.de"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private Destination getDestination(final UUID destinationId, final SubmissionSchema schema) {
        final var destinationService = new DestinationService();
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    public static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}