package de.fitko.fitconnect.core.crypto;

import de.fitko.fitconnect.api.exceptions.InitializationException;
import de.fitko.fitconnect.api.services.crypto.MessageDigestService;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

public class HashService implements MessageDigestService {

    public static final String DEFAULT_ALGORITHM = "SHA-512"; // Currently, only SHA-512 is supported.

    private final MessageDigest messageDigest;

    public HashService() {
        try {
            this.messageDigest = MessageDigest.getInstance(DEFAULT_ALGORITHM);
        } catch (final NoSuchAlgorithmException e) {
            throw new InitializationException(e.getMessage(), e);
        }
    }

    public HashService(final MessageDigest messageDigest) {
        this.messageDigest = messageDigest;
    }

    @Override
    public byte[] createHash(final byte[] data) {
        this.messageDigest.reset();
        return messageDigest.digest(data);
    }

    @Override
    public boolean verify(final byte[] originalHash, final byte[] data) {
        final byte[] newHash = createHash(data);
        return compareHashes(originalHash, newHash);
    }

    @Override
    public String toHexString(final byte[] hash) {
        final StringBuffer hexStringBuffer = new StringBuffer();
        for (int i = 0; i < hash.length; i++) {
            hexStringBuffer.append(byteToHex(hash[i]));
        }
        return hexStringBuffer.toString();
    }

    @Override
    public byte[] fromHexString(final String hexString) {
        if (hexString.length() % 2 == 1) {
            throw new IllegalArgumentException("Invalid hexadecimal String.");
        }

        final byte[] bytes = new byte[hexString.length() / 2];
        for (int i = 0; i < hexString.length(); i += 2) {
            bytes[i / 2] = hexToByte(hexString.substring(i, i + 2));
        }
        return bytes;
    }

    private byte hexToByte(final String hexString) {
        final int firstDigit = toDigit(hexString.charAt(0));
        final int secondDigit = toDigit(hexString.charAt(1));
        return (byte) ((firstDigit << 4) + secondDigit);
    }

    private int toDigit(final char hexChar) {
        final int digit = Character.digit(hexChar, 16);
        if(digit == -1) {
            throw new IllegalArgumentException("Invalid Hexadecimal Character: "+ hexChar);
        }
        return digit;
    }

    private String byteToHex(final byte num) {
        final char[] hexDigits = new char[2];
        hexDigits[0] = Character.forDigit((num >> 4) & 0xF, 16);
        hexDigits[1] = Character.forDigit((num & 0xF), 16);
        return new String(hexDigits);
    }

    private boolean compareHashes(final byte[] originalHash, final byte[] comparisonHash) {
        int diff = originalHash.length ^ comparisonHash.length;
        for (int i = 0; i < originalHash.length && i < comparisonHash.length; i++) {
            diff |= originalHash[i] ^ comparisonHash[i];
        }
        return diff == 0;
    }

}
