package de.fitko.fitconnect.core.events;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.JWSSigner;
import com.nimbusds.jose.crypto.RSASSASigner;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.KeyOperation;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import de.fitko.fitconnect.api.exceptions.EventCreationException;
import de.fitko.fitconnect.api.exceptions.InvalidSigningKeyException;
import de.fitko.fitconnect.api.services.events.SecurityEventService;

import java.text.ParseException;
import java.util.Date;
import java.util.Map;
import java.util.UUID;

public class SecurityEventTokenService implements SecurityEventService {

    static final String ACCEPT_EVENT = "https://schema.fitko.de/fit-connect/events/accept-submission";
    static final String REJECT_EVENT = "https://schema.fitko.de/fit-connect/events/reject-submission";

    private final RSAKey signingKey;

    public SecurityEventTokenService(final String signingKey) {
        try {
            final var jwk = JWK.parse(signingKey).toJSONObject();
            this.signingKey = RSAKey.parse(jwk);
        } catch (final ParseException e) {
            throw new InvalidSigningKeyException("Signing key could not be parsed", e);
        }
    }

    @Override
    public String createAcceptSubmissionEvent(final UUID submissionId, final UUID destinationId, final UUID caseId) {

        validateTrueOrElseThrow(signingKey != null, "Signing key must not be null");

        final String transactionId = "case:" + caseId;
        final String subject = "submission:" + submissionId;

        validateRSAKey(signingKey);

        final JWSSigner signer = getJwsSigner(signingKey);
        final JWTClaimsSet claimsSet = getJwtClaimsSet(ACCEPT_EVENT, destinationId.toString(),transactionId, subject);
        final JWSHeader header = getJwsHeader(signingKey);
        final SignedJWT signedJWT = new SignedJWT(header, claimsSet);

        try {
            signedJWT.sign(signer);
            return signedJWT.serialize();
        } catch (final JOSEException e) {
            throw new EventCreationException("Signing confirmation JWT failed", e);
        }
    }

    private JWSSigner getJwsSigner(final RSAKey rsaKey) {
        try {
            return new RSASSASigner(rsaKey);
        } catch (final JOSEException e) {
            throw new EventCreationException("Error creating the RSASigner", e);
        }
    }


    private JWSHeader getJwsHeader(final JWK signingKey) {
        try{
            return JWSHeader.parse(Map.of(
                    "typ", "secevent+jwt",
                    "kid", signingKey.getKeyID(),
                    "alg", "PS512"
            ));
        }catch(final ParseException e){
            throw new EventCreationException("Parsing JWS header failed", e);
        }
    }

    private JWTClaimsSet getJwtClaimsSet(final String event, final String destinationId, final String transactionId, final String subject) {
        return new JWTClaimsSet.Builder()
                .issuer(destinationId)
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject(subject)
                .claim("events", Map.of(event, Map.of()))
                .claim("txn", transactionId)
                .build();
    }

    @Override
    public String createRejectSubmissionEvent(final UUID submissionId, final UUID destinationId, final UUID caseId) {
        validateTrueOrElseThrow(signingKey != null, "Signing key must not be null");

        final String transactionId = "case:" + caseId;
        final String subject = "submission:" + submissionId;

        validateRSAKey(signingKey);

        final JWSSigner signer = getJwsSigner(signingKey);
        final JWTClaimsSet claimsSet = getJwtClaimsSet(REJECT_EVENT, destinationId.toString(), transactionId, subject);
        final JWSHeader header = getJwsHeader(signingKey);
        final SignedJWT signedJWT = new SignedJWT(header, claimsSet);

        try {
            signedJWT.sign(signer);
            return signedJWT.serialize();
        } catch (final JOSEException e) {
            throw new EventCreationException("Signing rejection JWT failed", e);
        }
    }

    public static void validateRSAKey(final RSAKey RSAKey){
        validateTrueOrElseThrow(RSAKey.getModulus().decodeToBigInteger().bitLength() >= 4096, "JWK has wrong key length.");
        validateTrueOrElseThrow(RSAKey.getAlgorithm().equals(JWSAlgorithm.PS512), "The specified public key does not use PS512 as algorithm.");
        //validateTrueOrElseThrow(RSAKey.getKeyOperations().size() == 1 && RSAKey.getKeyOperations().contains(KeyOperation.VERIFY), "The specified public key is not intended for 'verify' as specified through key operation.");
        validateTrueOrElseThrow(RSAKey.getPublicExponent().toString().equals("AQAB"), "The specified public key does not match the public exponent 'AQAB'.");

    }

    private static void validateTrueOrElseThrow(final boolean expression, final String msg) {
        if (!expression) {
            throw new EventCreationException(msg);
        }
    }



}
