package de.fitko.fitconnect.core.validation;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.networknt.schema.JsonSchema;
import com.networknt.schema.JsonSchemaFactory;
import com.networknt.schema.SpecVersion;
import com.networknt.schema.ValidationMessage;
import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.DataIntegrityException;
import de.fitko.fitconnect.api.exceptions.ValidationException;
import de.fitko.fitconnect.api.services.crypto.MessageDigestService;
import de.fitko.fitconnect.api.services.validation.ValidationService;

import java.util.Set;
import java.util.stream.Collectors;

public class DefaultValidationService implements ValidationService {

    private final MessageDigestService messageDigestService;

    public DefaultValidationService(final MessageDigestService messageDigestService) {
        this.messageDigestService = messageDigestService;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        // TODO: Call to external Lib
        //  Will be implemented in dedicated validation lib from Zustelldienst
        //  That lib is still under development and its API should be used here to call the actual validation
        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateMetadataSchema(final String jsonSchema, final Metadata metadata) {
        try {
            final ObjectMapper mapper = new ObjectMapper();
            final String metadataJson = mapper.writeValueAsString(metadata);
            final JsonNode schemaNode = mapper.readTree(jsonSchema);
            final JsonNode inputNode = mapper.readTree(metadataJson);
            final JsonSchemaFactory factory = JsonSchemaFactory.getInstance(SpecVersion.VersionFlag.V202012);
            final JsonSchema schema = factory.getSchema(schemaNode);
            final Set<ValidationMessage> errors = schema.validate(inputNode);
            return returnValidationResult(errors);
        } catch (final JsonProcessingException e) {
            return ValidationResult.error(new ValidationException(e.getMessage(), e));
        }
    }

    private ValidationResult returnValidationResult(final Set<ValidationMessage> errors) {
        if (errors.isEmpty()) {
            return ValidationResult.ok();
        }
        final String validationErrors = errors.stream()
                .map(ValidationMessage::getMessage)
                .collect(Collectors.joining("\n"));
        return ValidationResult.error(new ValidationException(validationErrors));
    }

    @Override
    public ValidationResult validateIntegrity(final String originalHexHash, final byte[] data) {
        final byte[] originalHash = messageDigestService.fromHexString(originalHexHash);
        final boolean hashesAreEqual = messageDigestService.verify(originalHash, data);
        return hashesAreEqual ? ValidationResult.ok() : ValidationResult.error(new DataIntegrityException(
                "Hash sum of transmitted data does not equal the hash of the sender."));
    }
}
