package de.fitko.fitconnect.core;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.DataIntegrityException;
import de.fitko.fitconnect.api.exceptions.RestApiException;
import de.fitko.fitconnect.api.services.Subscriber;
import de.fitko.fitconnect.api.services.crypto.CryptoService;
import de.fitko.fitconnect.api.services.events.EventLogService;
import de.fitko.fitconnect.api.services.events.SecurityEventService;
import de.fitko.fitconnect.api.services.submission.SubmissionService;
import de.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyInt;
import static org.mockito.Mockito.*;

public class SubmissionSubscriberTest {

    private SubmissionService submissionServiceMock;
    private EventLogService eventLogServiceMock;
    private CryptoService cryptoServiceMock;
    private ValidationService validationServiceMock;
    private SecurityEventService securityEventServiceMock;

    private Subscriber underTest;

    @BeforeEach
    public void setUp() {

        submissionServiceMock = Mockito.mock(SubmissionService.class);
        eventLogServiceMock = Mockito.mock(EventLogService.class);
        cryptoServiceMock = Mockito.mock(CryptoService.class);
        validationServiceMock = Mockito.mock(ValidationService.class);
        securityEventServiceMock = Mockito.mock(SecurityEventService.class);

        underTest = new SubmissionSubscriber(
                submissionServiceMock,
                eventLogServiceMock,
                cryptoServiceMock,
                validationServiceMock,
                securityEventServiceMock
        );
    }

    @Test
    public void decryptStringContent() throws JOSEException {

        // Given
        final RSAKey privateKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();

        final var expectedDecryptedString = "decryptedString";

        when(cryptoServiceMock.decryptBytes(any(),any())).thenReturn(expectedDecryptedString.getBytes());

        // When
        final byte[] decryptedBytes = underTest.decryptStringContent(privateKey, "encryptedString");

        // Then
        assertThat(new String(decryptedBytes), equalTo(expectedDecryptedString));
    }

    @Test
    public void pollAvailableSubmissions() {

        // Given
        final var destinationId = UUID.randomUUID();

        final SubmissionForPickup submission = new SubmissionForPickup();
        final SubmissionsForPickup availableSubmissions = new SubmissionsForPickup();
        availableSubmissions.setSubmissions(Set.of(submission));

        when(submissionServiceMock.pollAvailableSubmissions(any(),anyInt(), anyInt())).thenReturn(availableSubmissions);

        // When
        final Set<SubmissionForPickup> submissionForPickups = underTest.pollAvailableSubmissions(destinationId,0,100);

        // Then
        assertThat(submissionForPickups, equalTo(availableSubmissions.getSubmissions()));
        assertThat(submissionForPickups.size(), is(availableSubmissions.getSubmissions().size()));
    }

    @Test
    public void getSubmission() {

        // Given
        final var submissionId = UUID.randomUUID();

        final var expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submissionId);

        when(submissionServiceMock.getSubmission(any())).thenReturn(expectedSubmission);

        // When
        final Submission submission = underTest.getSubmission(submissionId);

        // Then
        assertNotNull(submission);
        assertThat(submission.getSubmissionId(), is(submissionId));
    }

    @Test
    public void submissionNotFound() {

        // Given
        when(submissionServiceMock.getSubmission(any())).thenThrow(RestApiException.class);

        // Then
        assertThrows(
                RestApiException.class,
                () ->  underTest.getSubmission(UUID.randomUUID()),
                "Expected getSubmission() to throw, but nothing was thrown"
        );
    }

    @Test
    public void fetchAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var expectedAttachment = "encryptedAttachmentContent";

        when(submissionServiceMock.getAttachment(any(), any())).thenReturn(expectedAttachment);

        // When
        final String encryptedAttachment = underTest.fetchAttachment(submissionId, attachmentId);

        // Then
        assertThat(encryptedAttachment, equalTo(expectedAttachment));
    }

    @Disabled("Needs proper impl first")
    void validateEventLog() {

    }

    @Test
    public void validateMetadata() {

        // Given
        when(validationServiceMock.validateMetadataSchema(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateMetadata(new Metadata(), "jsonSchema");

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void invalidMetadata() {

        // Given
        when(validationServiceMock.validateMetadataSchema(any(), any())).thenReturn(ValidationResult.error(new Exception("Failed parsing metadata")));

        // When
        final ValidationResult validationResult = underTest.validateMetadata(new Metadata(), "jsonSchema");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Failed parsing metadata"));
    }

    @Test
    public void validateHashIntegrity() {

        when(validationServiceMock.validateIntegrity(any(),any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateHashIntegrity("originalHex", "data".getBytes());

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void validateCorruptHash() {

        // Given
        when(validationServiceMock.validateIntegrity(any(),any())).thenReturn(ValidationResult.error(
                new DataIntegrityException("Hash sum of transmitted data does not equal the hash of the sender.")));

        // When
        final ValidationResult validationResult = underTest.validateHashIntegrity("originalHexHash", "corruptData".getBytes());

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("data does not equal the hash of the sender."));
    }

    @Test
    public void confirmValidSubmission() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        final var confirmationEvent = "CONFIRMED_EVENT";

        when(securityEventServiceMock.createAcceptSubmissionEvent(any(), any(), any())).thenReturn(confirmationEvent);
        doNothing().when(eventLogServiceMock).sendEvent(any(), any());

        // When
        underTest.confirmValidSubmission(submissionId, destinationId, caseId);

        // Then
        verify(eventLogServiceMock, times(1)).sendEvent(caseId, confirmationEvent);
    }

    @Test
    public void rejectSubmission() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        final var rejectionEvent = "REJECTED_EVENT";

        when(securityEventServiceMock.createRejectSubmissionEvent(any(), any(), any())).thenReturn(rejectionEvent);
        doNothing().when(eventLogServiceMock).sendEvent(any(), any());

        // When
        underTest.rejectSubmission(submissionId, destinationId, caseId);

        // Then
        verify(eventLogServiceMock, times(1)).sendEvent(caseId, rejectionEvent);
    }
}