package de.fitko.fitconnect.core.destination;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.fitko.fitconnect.api.config.ApplicationConfig;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.destination.Destination;
import de.fitko.fitconnect.api.domain.model.jwk.JWK;
import de.fitko.fitconnect.api.services.auth.OAuthService;
import de.fitko.fitconnect.api.services.destination.DestinationService;
import de.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestTemplate;

import java.util.UUID;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class DestinationRestServiceTest extends RestEndpointBase {

    DestinationService underTest;
    OAuthService authServiceMock;

    @BeforeEach
    void setup() {
        final ApplicationConfig config = getTestConfig("http://localhost:" + wireMockServer.port());
        authServiceMock = mock(OAuthService.class);
        underTest = new DestinationRestService(authServiceMock, new RestTemplate(), config);
    }

    @Test
    public void getDestination() throws JsonProcessingException {

        //Given
        final Destination expectedDestination = new Destination();
        expectedDestination.setDestinationId(UUID.randomUUID());

        final OAuthToken oAuthToken = new OAuthToken();
        oAuthToken.setAccessToken("123");
        when(authServiceMock.getCurrentToken()).thenReturn(oAuthToken);

        final var urlPath = "/v1/destinations/" + expectedDestination.getDestinationId();
        stubFor(urlPath, expectedDestination, MediaType.APPLICATION_JSON_VALUE);

        // When
        final Destination destination = underTest.getDestination(expectedDestination.getDestinationId());

        // Then
        assertEquals(expectedDestination, destination);
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer 123")));

    }

    @Test
    public void getEncryptionKey() throws Exception {
        // Given
        final UUID destinationId = UUID.randomUUID();
        final JWK jwk = new JWK();
        jwk.setKid("123");

        final OAuthToken oAuthToken = new OAuthToken();
        oAuthToken.setAccessToken("123");
        when(authServiceMock.getCurrentToken()).thenReturn(oAuthToken);

        final var urlPath = "/v1/destinations/" + destinationId + "/keys/" + jwk.getKid();
        stubFor(urlPath, jwk, MediaType.APPLICATION_JSON_VALUE);

        // When
        final JWK publicKey = underTest.getEncryptionKey(destinationId, jwk.getKid());

        // Then
        assertEquals(jwk, publicKey);
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer 123")));
    }

    private void stubFor(final String url, final Object responseBody, final String contentType) throws JsonProcessingException {
        wireMockServer.stubFor(
                get(urlEqualTo(url))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(responseBody))
                                .withHeader("Content-Type", contentType)
                                .withStatus(200)));
    }

}