package de.fitko.fitconnect.core.validation;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import de.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.data.*;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.services.crypto.MessageDigestService;
import de.fitko.fitconnect.api.services.validation.ValidationService;
import de.fitko.fitconnect.core.crypto.HashService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.net.URI;
import java.util.Collections;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class DefaultValidationServiceTest {

    ValidationService underTest;
    MessageDigestService hashService;

    @BeforeEach
    void setUp() {
        hashService = new HashService();
        underTest = new DefaultValidationService(hashService);
    }

    @Test
    public void validatePublicKey() throws JOSEException {

        //Given
        final RSAKey rsaKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();

        // When
        final ValidationResult validationResult = underTest.validatePublicKey(rsaKey);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void validateMetadata() throws IOException {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent(hashService.toHexString(hashService.createHash("someBogusContent".getBytes())));
        hash.setDataSignatureType(DataSignatureType.SHA_512);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String schema = getSchema("/metadata_schema.json");

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(schema, metadata);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void validateMetadataWithMissingProperties() throws IOException {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var dataWithoutHashedContent = new Data();
        dataWithoutHashedContent.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(dataWithoutHashedContent);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String schema = getSchema("/metadata_schema.json");

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(schema, metadata);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$.contentStructure.data.hash: is missing but it is required"));
    }

    @Test
    public void validateMetadataWithInvalidPattern() throws IOException {

        // Given
        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var hash = new Hash();
        hash.setContent("noHexString");
        hash.setDataSignatureType(DataSignatureType.SHA_512);

        final var dataWithInvalidHashFormat = new Data();
        dataWithInvalidHashFormat.setSubmissionSchema(submissionSchema);
        dataWithInvalidHashFormat.setHash(hash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(dataWithInvalidHashFormat);
        contentStructure.setAttachments(Collections.emptyList());

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String schema = getSchema("/metadata_schema.json");

        // When
        final ValidationResult validationResult = underTest.validateMetadataSchema(schema, metadata);

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("$.contentStructure.data.hash.content: does not match the regex pattern ^[a-f0-9]{128}$"));
    }

    @Test
    public void validateMatchingHash() {

        // Given
        final var hexHashFromSender = hashService.toHexString(hashService.createHash("some data".getBytes()));

        // When
        final ValidationResult validationResult = underTest.validateIntegrity(hexHashFromSender, "some data".getBytes());

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    public void validateCorruptHash() {

        // Given
        final var hexHashFromSender = "e1645e7492f032fb62c674db75500be7b260bfc0daa965821ddb3f8a49b5d33788ee3f04";

        // When
        final ValidationResult validationResult = underTest.validateIntegrity(hexHashFromSender, "corrupted data".getBytes());

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("transmitted data does not equal the hash of the sender"));
    }

    private String getSchema(final String filename) throws IOException {
        return new String(DefaultValidationServiceTest.class.getResourceAsStream(filename).readAllBytes());
    }
}